/*
 * Driver for the ColdFire Flash Module on MCF5282
 * 
 * Author: Mathias Kster
 * Modified for u-boot by Zachary P. Landau
 *
 * Copyright (c) 2004, Mathias Kster
 *
 * Some parts are based on lart.c by Abraham vd Merwe
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

/* extra verbose debugging */
#undef CFM_DEBUG

#include <common.h>
#include <watchdog.h>

#ifdef CFG_FLASH_CFM_DRIVER

#include <asm/m5282.h>

#ifndef CONFIG_M5282
#error This is for MCF5282 architecture only
#endif

typedef unsigned long flash_sect_t;
extern flash_info_t flash_info[CFG_MAX_FLASH_BANKS];

#define CFM_FLASH_ID			0xCF5282

/*
 * Flash memory blocksize and blockcount.
 */
#define CFM_FLASH_BLOCKSIZE		(1024*2)
#define CFM_FLASH_NUMBLOCKS		256

/*
 * General settings, FLASH_OFFSET based on FLASHBAR chip select
 */
#define CFM_BUSWIDTH		4
#define CFM_OFFSET		CFG_INT_FLASH_BASE
#define CFM_BACKDOOR_OFFSET	(CFG_MBAR+0x04000000)

/*
 * CFM User Mode Commands
 */
#define CFMCMD_RDARY1	0x05	/* erase verify */
#define CFMCMD_PGM	0x20	/* longword program */
#define CFMCMD_PGERS	0x40	/* page erase */
#define CFMCMD_MASERS	0x41	/* mass erase */
#define CFMCMD_PGERSVER	0x06	/* page erase verify */

/*
 * CFM Status Bits
 */
#define CFMUSTAT_CBEIF	0x80
#define CFMUSTAT_CCIF	0x40
#define CFMUSTAT_PVIOL	0x20
#define CFMUSTAT_ACCERR	0x10
#define CFMUSTAT_BLANK	0x04

/***************************************************************************************************/

/*
 * Reset CFM state register error bits
 */
void reset_state (void)
{
   if ( *(volatile unsigned char*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_PVIOL )
		*(volatile unsigned char*) (CFG_MBAR+MCFCFM_USTAT) |= CFMUSTAT_PVIOL;
   if ( *(volatile unsigned char*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_ACCERR )
		*(volatile unsigned char*) (CFG_MBAR+MCFCFM_USTAT) |= CFMUSTAT_ACCERR;
   if ( *(volatile unsigned char*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_BLANK )
		*(volatile unsigned char*) (CFG_MBAR+MCFCFM_USTAT) |= CFMUSTAT_BLANK;
}

/*
 * Erase one block of flash memory at offset ``offset'' which is any
 * address within the block which should be erased.
 *
 * Returns 1 if successful, 0 otherwise.
 */
static inline int cfm_erase_block (__u32 offset)
{
   __u32 status = 1;

   debug("real offset: 0x%.8x\n", offset);
   offset = offset - CFM_OFFSET;

#ifdef DEBUG
   debug("\n%s(): 0x%.8x\n",__FUNCTION__,offset);
#endif

   /* start sector erase sequence */
   *(volatile __u32*) (CFM_BACKDOOR_OFFSET+offset) = 0;
   *(volatile __u8*) (CFG_MBAR+MCFCFM_CMD) = CFMCMD_PGERS;
   *(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) = CFMUSTAT_CBEIF;
   
   /* wait for ready */
   while ( ! (*(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_CCIF ) );
   
   /* error handling */
   if ( *(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_ACCERR )
   	{
   		debug( "%s(): ACCERR\n",__FUNCTION__);
		status = 0;
   	}
   if ( *(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_PVIOL )
   	{
   		debug( "%s(): PVIOL\n",__FUNCTION__);
		status = 0;
   	}

   /* reset status flags */
   reset_state();

   return (status);
}

int cfm_flash_erase (flash_info_t *info, int s_first, int s_last)
{
	flash_sect_t sect;
	int prot;

#ifdef DEBUG
   debug ("\n%s(): 0x%.8x-0x%.8x\n",__FUNCTION__,s_first,s_last);
#endif

	prot = 0;
	for (sect = s_first; sect <= s_last; ++sect) {
		if (info->protect[sect])
			prot++;
	}
	if (prot)
		printf("Warning: %d protected sectors will not be erased!\n", prot);

	for (sect = s_first; sect <= s_last; sect++) {
		if (info->protect[sect] == 0)
			cfm_erase_block(sect*CFM_FLASH_BLOCKSIZE);
		WATCHDOG_RESET();
		putc(".");
	}
	puts (" done\n");
	return 0;
}

#if 0
static int flash_erase_blah (struct mtd_info *mtd,struct erase_info *instr)
{
   __u32 addr,len;
   int i,first;

#ifdef DEBUG
   debug("%s(addr = 0x%.8x, len = %d)\n",__FUNCTION__,instr->addr,instr->len);
#endif

   /* sanity checks */
   if (instr->addr + instr->len > mtd->size) return (-EINVAL);

   /*
	* check that both start and end of the requested erase are
	* aligned with the erasesize at the appropriate addresses.
	*
	* skip all erase regions which are ended before the start of
	* the requested erase. Actually, to save on the calculations,
	* we skip to the first erase region which starts after the
	* start of the requested erase, and then go back one.
	*/
   for (i = 0; i < mtd->numeraseregions && instr->addr >= mtd->eraseregions[i].offset; i++) ;
   i--;

   /*
	* ok, now i is pointing at the erase region in which this
	* erase request starts. Check the start of the requested
	* erase range is aligned with the erase size which is in
	* effect here.
	*/
   if (instr->addr & (mtd->eraseregions[i].erasesize - 1)) return (-EINVAL);

   /* Remember the erase region we start on */
   first = i;

   /*
	* next, check that the end of the requested erase is aligned
	* with the erase region at that address.
	*
	* as before, drop back one to point at the region in which
	* the address actually falls
	*/
   for (; i < mtd->numeraseregions && instr->addr + instr->len >= mtd->eraseregions[i].offset; i++) ;
   i--;

   /* is the end aligned on a block boundary? */
   if ((instr->addr + instr->len) & (mtd->eraseregions[i].erasesize - 1)) return (-EINVAL);

   addr = instr->addr;
   len = instr->len;

   i = first;

   /* now erase those blocks */
   while (len)
	 {
		if (!cfm_erase_block (addr))
		  {
			 instr->state = MTD_ERASE_FAILED;
			 return (-EIO);
		  }

		addr += mtd->eraseregions[i].erasesize;
		len -= mtd->eraseregions[i].erasesize;

		if (addr == mtd->eraseregions[i].offset + (mtd->eraseregions[i].erasesize * mtd->eraseregions[i].numblocks)) i++;
	 }

   instr->state = MTD_ERASE_DONE;
   if (instr->callback) instr->callback (instr);

   return (0);
}

static int flash_read (struct mtd_info *mtd,loff_t from,size_t len,size_t *retlen,u_char *buf)
{
#ifdef DEBUG
   debug ("%s(from = 0x%.8x, len = %d)\n",__FUNCTION__,(__u32) from,len);
#endif

   /* sanity checks */
   if (!len) return (0);
   if (from + len > mtd->size) return (-EINVAL);

   /* we always read len bytes */
   *retlen = len;

   memcpy(buf,((u_char*)(CFM_OFFSET))+from,len);
   
   return 0;
}
#endif

/*
 * Write one dword ``x'' to flash memory at offset ``offset''. ``offset''
 * must be 32 bits, i.e. it must be on a dword boundary.
 *
 * Returns 1 if successful, 0 otherwise.
 */
static inline int write_dword (__u32 offset,__u32 x)
{
   __u32 status = 1;

#ifdef CFM_DEBUG
   debug ("%s(): 0x%.8x <- 0x%.8x\n",__FUNCTION__,offset,x);
#endif

   /* start program sequence */
   *(volatile __u32*) (CFM_BACKDOOR_OFFSET+offset) = x;
   *(volatile __u8*) (CFG_MBAR+MCFCFM_CMD) = CFMCMD_PGM;
   *(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) = CFMUSTAT_CBEIF;

   /* wait for ready */
   while ( ! (*(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_CCIF ) );
   
   /* error handling */
   if ( *(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_ACCERR )
   	{
   		debug( "%s(): ACCERR\n",__FUNCTION__);
		status = 0;
   	}
   if ( *(volatile __u8*) (CFG_MBAR+MCFCFM_USTAT) & CFMUSTAT_PVIOL )
   	{
   		debug( "%s(): PVIOL\n",__FUNCTION__);
		status = 0;
   	}

   /* reset status flags */
   reset_state();
   
   return status;
}

int cfm_flash_write (flash_info_t *info, uchar *buf, ulong to, ulong len)
{
   __u8 tmp[4];
   int i,n;

   to = to - CFM_OFFSET;

#ifdef DEBUG
   debug ("%s(to = 0x%.8x, len = %ld)\n",__FUNCTION__,(__u32) to,len);
#endif

   /* sanity checks */
   if (!len) return (0);
   if (to + len > info->size) return (-1);

   /* first, we write a 0xFF.... padded byte until we reach a dword boundary */
   if (to & (CFM_BUSWIDTH - 1))
	 {
		__u32 aligned = to & ~(CFM_BUSWIDTH - 1);
		int gap = to - aligned;

		i = n = 0;

		while (gap--) tmp[i++] = 0xFF;
		while (len && i < CFM_BUSWIDTH) tmp[i++] = buf[n++], len--;
		while (i < CFM_BUSWIDTH) tmp[i++] = 0xFF;

		if (!write_dword (aligned,*((__u32 *) tmp))) return (-1);

		to += n;
		buf += n;
	 }

   /* now we write dwords until we reach a non-dword boundary */
   while (len >= CFM_BUSWIDTH)
	 {
		if (!write_dword (to,*((__u32 *) buf))) return (-1);

		to += CFM_BUSWIDTH;
		buf += CFM_BUSWIDTH;
		len -= CFM_BUSWIDTH;
	 }

   /* top up the last unaligned bytes, padded with 0xFF.... */
   if (len & (CFM_BUSWIDTH - 1))
	 {
		i = n = 0;

		while (len--) tmp[i++] = buf[n++];
		while (i < CFM_BUSWIDTH) tmp[i++] = 0xFF;

		if (!write_dword (to,*((__u32 *) tmp))) return (-1);
	 }

   return (0);
}

void cfm_print_info(flash_info_t *info)
{
	int i;
	if (info->flash_id != CFM_FLASH_ID) {
		puts ("Invalid FLASH type\n");
		return;
	}

	printf ("CFM FLASH: %uK in %ld sectors)\n",
			info->size / 1024, info->sector_count);
	puts ("  Sector Start Addresses:");
	for (i=0; i < info->sector_count; ++i) {
		if ((i % 5) == 0)
			printf ("\n   ");
		printf (" %08lX%s",
			info->start[i], info->protect[i] ? " (RO)  " : "     ");
	}
	putc ('\n');
	return;
}

/***************************************************************************************************/

#define NB_OF(x) (sizeof (x) / sizeof (x[0]))

unsigned long cfm_flash_init (int bank)
{
   unsigned long size;
   unsigned long sector;
   int i;
   //debug ("MTD driver for ColdFire Flash Module. Written by Mathias Kster\n");

   reset_state();

   *(volatile unsigned short*)(CFG_MBAR+MCFCFM_MCR)  = 0;
   /* 200KHz @ 64MHz sysclk */
   *(volatile unsigned char*) (CFG_MBAR+MCFCFM_CLKD) = (1<<6) | 20;
   /* disable protection */
   *(volatile unsigned long*) (CFG_MBAR+MCFCFM_PROT) = 0;
   /* disable supervisor mode */
   *(volatile unsigned long*) (CFG_MBAR+MCFCFM_SACC) = 0;
   /* data and program address space */
   *(volatile unsigned long*) (CFG_MBAR+MCFCFM_DACC) = 0;

   if ( (*(volatile unsigned long*) (CFG_MBAR+MCFCFM_SEC) & 0x80000000) )
   	debug ("CFM: Warning, backdoor mode enabled. (%lx)\n",*(volatile unsigned long*) (CFG_MBAR+MCFCFM_SEC));
   if ( (*(volatile unsigned long*) (CFG_MBAR+MCFCFM_SEC) & 0x40000000) )
   	debug ("CFM: Warning, security mode enabled. %lx\n",*(volatile unsigned long*) (CFG_MBAR+MCFCFM_SEC));


   flash_info[bank].flash_id = CFM_FLASH_ID;
   size = flash_info[bank].size = CFM_FLASH_BLOCKSIZE * CFM_FLASH_NUMBLOCKS;
   flash_info[bank].sector_count = CFM_FLASH_NUMBLOCKS;
   flash_info[bank].print_info = cfm_print_info;
   flash_info[bank].flash_erase = cfm_flash_erase;
   flash_info[bank].write_buff = cfm_flash_write;

#ifdef DEBUG
   debug (	   "flash_id = 0x%.8x\n"
		   "size = 0x%.8lx (%uK)\n"
		   "sector_size = 0x%.8lx (%uK)\n"
		   "sector_count = %ld\n",
		   flash_info[bank].flash_id,
		   flash_info[bank].size, flash_info[bank].size / 1024,
		   CFM_FLASH_BLOCKSIZE,CFM_FLASH_BLOCKSIZE / 1024,
		   flash_info[bank].sector_count);
#endif

   sector = CFM_OFFSET;
   for (i=0; i < flash_info[bank].sector_count; i++) {
	   flash_info[bank].start[i] = sector;
	   sector += CFM_FLASH_BLOCKSIZE;
	   flash_info[bank].protect[i] = 1;
   }

#if 0
   if (mtd.numeraseregions)
	 for (result = 0; result < mtd.numeraseregions; result++)
	   debug (
			   "\n\n"
			   "mtd.eraseregions[%d].offset = 0x%.8x\n"
			   "mtd.eraseregions[%d].erasesize = 0x%.8x (%uK)\n"
			   "mtd.eraseregions[%d].numblocks = %d\n",
			   result,mtd.eraseregions[result].offset,
			   result,mtd.eraseregions[result].erasesize,mtd.eraseregions[result].erasesize / 1024,
			   result,mtd.eraseregions[result].numblocks);

#endif

   return (size);
}

#endif /* CFG_FLASH_CFM_DRIVER */
